#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2011 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using System.Reflection;
using System.IO;
using HeuristicLab.Persistence.Default.Xml;
using System.Diagnostics;
using HeuristicLab.Optimization;
using System.Linq;

namespace HeuristicLab.MPIEngine {
  /// <summary>
  /// Represents an engine that executes its steps in parallel (if possible) using multiple threads.
  /// This engine is suitable for parallel processing on shared memory systems which provide multiple cores.
  /// </summary>
  [StorableClass]
  [Item("MPI Engine", "Engine for parallel execution of algorithms using multiple processes (suitable for distributed memory systems with multiple cores).")]
  public class MPIEngine : Engine {
    [StorableConstructor]
    protected MPIEngine(bool deserializing) : base(deserializing) { }
    protected MPIEngine(MPIEngine original, Cloner cloner) : base(original, cloner) { }
    public MPIEngine() : base() { }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new MPIEngine(this, cloner);
    }

    private string algFile;

    public override void Start() {
      if (ExecutionStack.Count == 1) {
        ExecutionContext context = ExecutionStack.First() as ExecutionContext;

        ExecutionContext algorithmContext = context.Parent as ExecutionContext;

        EngineAlgorithm alg = typeof(ExecutionContext).InvokeMember("parameterizedItem",
          BindingFlags.GetField | BindingFlags.NonPublic |
          BindingFlags.Instance, null, algorithmContext, null) as EngineAlgorithm;

        alg = alg.Clone() as EngineAlgorithm;
        alg.Engine = new SequentialEngine.SequentialEngine();

        algFile = Path.GetTempFileName();
        XmlGenerator.Serialize(alg, algFile);
      }

      base.Start();
    }

    protected override void Run(System.Threading.CancellationToken cancellationToken) {
      if (ExecutionStack.Count == 1) {
        ExecutionContext context = ExecutionStack.Pop() as ExecutionContext;

        IScope globalScope = context.Scope;

        string resultFile = Path.GetTempFileName();
        ItemList<ResultCollection> empty = new ItemList<ResultCollection>();
        XmlGenerator.Serialize(empty, resultFile);

        string exec = @"C:\Program Files\Microsoft Compute Cluster Pack\Bin\mpiexec.exe";
        string args = @"-n 3 HeuristicLab.MPIAlgorithmRunner-3.3.exe " + algFile + " " + resultFile + " 5000";

        System.Threading.Thread pollThread = new System.Threading.Thread(delegate(object state) {
          while (true) {
            System.Threading.Thread.Sleep(5000);

            lock (resultFile) {
              object results = XmlParser.Deserialize(resultFile);
              ResultCollection resultCollection = (globalScope.Variables["Results"].Value as ResultCollection);

              if (resultCollection != null) {
                if (!resultCollection.ContainsKey("MPIResults"))
                  resultCollection.Add(new Result("MPIResults", results as IItem));

                resultCollection["MPIResults"].Value = results as IItem;
              }

            }
          }
        });
        pollThread.Start();

        Process p = Process.Start(exec, args);
        p.WaitForExit();

        pollThread.Abort();
        File.Delete(algFile);
        File.Delete(resultFile);
      }
    }
  }
}
